# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import division, print_function

import numpy as np

__all__ = ()


def numerical_gradients(geom, func, *args, delta=1e-3, **kwargs):
    steps, diff = finite_diff_gen(kwargs.pop('npts', 5))
    coords_0 = geom.coords
    gradients = np.zeros(coords_0.shape)
    for i_atom in range(coords_0.shape[0]):
        for i_xyz in range(3):
            ene = {}
            for step in steps:
                coords = coords_0.copy()
                coords[i_atom, i_xyz] += step * delta
                geom.coords = coords
                ene[step] = getattr(geom, func)(*args, **kwargs)
            gradients[i_atom, i_xyz] = diff(ene, delta)
    return gradients


def numerical_latt_gradients(geom, func, *args, **kwargs):
    delta = kwargs.pop('delta', 1e-3)
    steps, diff = finite_diff_gen(kwargs.pop('npts', 5))
    lattice_0 = geom.lattice
    gradients = np.zeros((3, 3))
    for i_vec in range(3):
        for i_xyz in range(3):
            ene = {}
            for step in steps:
                lattice = lattice_0.copy()
                lattice[i_vec, i_xyz] += step * delta
                geom.lattice = lattice
                ene[step] = getattr(geom, func)(*args, **kwargs)
            gradients[i_vec, i_xyz] = diff(ene, delta)
    return gradients


def _diff3(x, delta):
    return (-1.0 / 2 * x[-1] + 1.0 / 2 * x[1]) / delta


def _diff5(x, delta):
    return (
        1.0 / 12 * x[-2] - 2.0 / 3 * x[-1] + 2.0 / 3 * x[1] - 1.0 / 12 * x[2]
    ) / delta


def finite_diff_gen(npts):
    return {3: ([-1, 1], _diff3), 5: ([-2, -1, 1, 2], _diff5)}[npts]



import os
import sys

import cffi

MBD_H = 'src/mbd.h'
LIBMBD_PREFIX = os.environ.get('LIBMBD_PREFIX')

if LIBMBD_PREFIX == '':
    sys.exit()

# some Conda environments do not add their include dir into default includes
CONDA_PREFIX = os.environ.get('CONDA_PREFIX')
if not LIBMBD_PREFIX and CONDA_PREFIX:
    LIBMBD_PREFIX = CONDA_PREFIX

if LIBMBD_PREFIX:
    ext_kwargs = {
        'include_dirs': [f'{LIBMBD_PREFIX}/include'],
        'library_dirs': [f'{LIBMBD_PREFIX}/lib'],
        'runtime_library_dirs': [f'{LIBMBD_PREFIX}/lib'],
    }
else:
    ext_kwargs = {}
ffibuilder = cffi.FFI()
ffibuilder.set_source(
    'pymbd._libmbd', '#include "mbd/mbd.h"', libraries=['mbd'], **ext_kwargs
)
with open(MBD_H) as f:
    ffibuilder.cdef(f.read())

if __name__ == '__main__':
    from cffi.setuptools_ext import cffi_modules
    from setuptools import Distribution

    if sys.platform[:6] == 'darwin':
        from distutils.unixccompiler import UnixCCompiler

        UnixCCompiler.runtime_library_dir_option = lambda self, dir: ['-rpath', dir]

    distribution = Distribution({'package_dir': {'': 'src'}})
    cffi_modules(distribution, 'cffi_modules', ['build.py:ffibuilder'])
    cmd = distribution.cmdclass['build_ext'](distribution)
    cmd.inplace = 1
    cmd.ensure_finalized()
    cmd.run()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import subprocess
import sys

import pytest

from pymbd.benchmark import parse


@pytest.mark.parametrize(
    'args', [[], ['--finite'], ['--method=ts']], ids=lambda x: ' '.join(x)
)
@pytest.mark.no_scalapack
def test_benchmark(args):
    stdout = subprocess.run(
        [sys.executable, '-u', '-m', 'pymbd.benchmark', *args],
        check=True,
        stdout=subprocess.PIPE,
    ).stdout.decode()
    assert parse(stdout)['energy']



import pytest
from pytest import approx

from pymbd import mbd_energy_species


@pytest.mark.no_scalapack
def test_ethylcarbamate(ethylcarbamate):
    enes = [
        mbd_energy_species(
            coords, species, vol_ratios, 0.83, lattice=lattice, k_grid=k_grid
        )
        for coords, lattice, k_grid, species, vol_ratios in ethylcarbamate
    ]
    ene_int = enes[0] - 2 * enes[1]
    assert ene_int == approx(-0.037040868610822564, rel=1e-10)



import numpy as np
import pytest
from pytest import approx

from pymbd import ang, from_volumes
from pymbd.fortran import MBDFortranError, MBDGeom
from pymbd.utils import numerical_gradients, numerical_latt_gradients


def test_argon_dimer_plain():
    ene = MBDGeom([(0, 0, 0), (0, 0, 4 * ang)]).mbd_energy(
        [11, 11], [63.525, 63.525], [3.55, 3.55], 0.83, variant='plain'
    )
    assert ene == approx(-0.00024329110270970844, rel=1e-10)


@pytest.mark.no_scalapack
def test_argon_dimer_dipole_matrix():
    dip = MBDGeom([(0, 0, 0), (0, 0, 4 * ang)]).dipole_matrix('bare')
    assert (dip != 0).sum() == 6


def test_argon_dimer_rsscs():
    ene = MBDGeom([(0, 0, 0), (0, 0, 4 * ang)]).mbd_energy_species(
        ['Ar', 'Ar'], [1, 1], 0.83
    )
    assert ene == approx(-0.0002462647623815428, rel=1e-10)


def test_argon_dimer_rsscs_rpa():
    geom = MBDGeom([(0, 0, 0), (0, 0, 4 * ang)], do_rpa=True, get_rpa_orders=True)
    ene, orders = geom.mbd_energy_species(['Ar', 'Ar'], [1, 1], 0.83)
    assert ene == approx(-0.0002462647623815428, rel=1e-10)
    assert orders[1] == approx(-0.0002461558113413099, rel=1e-10)
    assert orders[2] == approx(0)
    assert orders[3] == approx(-1.0885208380438466e-07, rel=1e-10)


def test_argon_dimer_ts():
    ene = MBDGeom([(0, 0, 0), (0, 0, 4 * ang)]).ts_energy(
        [11, 11], [63.525, 63.525], [3.55, 3.55], 0.94
    )
    assert ene == approx(-0.000318123017869182, rel=1e-10)


def test_benzene_dimer(benzene_dimer):
    mon1, mon2 = benzene_dimer
    dim = (np.vstack((mon1[0], mon2[0])), mon1[1] + mon2[1], mon1[2] + mon2[2])
    enes = [
        MBDGeom(coords).mbd_energy_species(species, vol_ratios, 0.83)
        for coords, species, vol_ratios in (mon1, mon2, dim)
    ]
    ene_int = enes[2] - enes[1] - enes[0]
    assert ene_int == approx(-0.006312323931302544, rel=1e-10)


def test_benzene_gradients(benzene_dimer):
    coords, species, vol_ratios = benzene_dimer[0]
    ene, gradients = MBDGeom(coords).mbd_energy_species(
        species, vol_ratios, 0.83, force=True
    )
    with MBDGeom(coords) as geom:
        num_gradients = numerical_gradients(
            geom, 'mbd_energy_species', species, vol_ratios, 0.83
        )
    for i in range(len(coords)):
        assert gradients[i] == approx(num_gradients[i], rel=1e-10, abs=1e-10)


@pytest.mark.no_scalapack
def test_benzene_dimer_python(benzene_dimer):
    mon1, mon2 = benzene_dimer
    dim = (np.vstack((mon1[0], mon2[0])), mon1[1] + mon2[1], mon1[2] + mon2[2])
    enes = [
        MBDGeom(coords).mbd_energy_species(species, vol_ratios, 0.83)
        for coords, species, vol_ratios in (mon1, mon2, dim)
    ]
    ene_int = enes[2] - enes[1] - enes[0]
    assert ene_int == approx(-0.006312323931302544, rel=1e-10)


def test_benzene_gradients_plain(benzene_dimer):
    coords, species, vol_ratios = benzene_dimer[0]
    ene, gradients = MBDGeom(coords).mbd_energy_species(
        species, vol_ratios, 0.83, variant='plain', force=True
    )
    with MBDGeom(coords) as geom:
        num_gradients = numerical_gradients(
            geom, 'mbd_energy_species', species, vol_ratios, 0.83, variant='plain'
        )
    for i in range(len(coords)):
        assert gradients[i] == approx(num_gradients[i], rel=1e-10, abs=1e-10)


def test_benzene_dimer_scs(benzene_dimer):
    mon1, mon2 = benzene_dimer
    dim = (np.vstack((mon1[0], mon2[0])), mon1[1] + mon2[1], mon1[2] + mon2[2])
    enes = [
        MBDGeom(coords).mbd_energy_species(
            species, vol_ratios, 1, a=2.56, variant='scs'
        )
        for coords, species, vol_ratios in (mon1, mon2, dim)
    ]
    ene_int = enes[2] - enes[1] - enes[0]
    assert ene_int == approx(-0.007462380657774048, rel=1e-10)


def test_benzene_dimer_ts(benzene_dimer):
    mon1, mon2 = benzene_dimer
    dim = (np.vstack((mon1[0], mon2[0])), mon1[1] + mon2[1], mon1[2] + mon2[2])
    enes = [
        MBDGeom(coords).ts_energy_species(species, vol_ratios, 0.94)
        for coords, species, vol_ratios in (mon1, mon2, dim)
    ]
    ene_int = enes[2] - enes[1] - enes[0]
    assert ene_int == approx(-0.008490052683234028, rel=1e-10)


def test_benzene(benzene_dimer):
    coords, species, vol_ratios = benzene_dimer[0]
    alpha_0, C6, R_vdw = from_volumes(species, vol_ratios)
    ene = MBDGeom(coords).mbd_energy(alpha_0, C6, R_vdw, 0.83, variant='plain')
    assert ene == approx(-0.007002398506090302, rel=1e-10)


def test_benzene_rpa(benzene_dimer):
    coords, species, vol_ratios = benzene_dimer[0]
    alpha_0, C6, R_vdw = from_volumes(species, vol_ratios)
    ene = MBDGeom(coords, do_rpa=True).mbd_energy(
        alpha_0, C6, R_vdw, 0.83, variant='plain'
    )
    assert ene == approx(-0.007002398506090302, rel=1e-9)


def test_benzene_rpa_scaled(benzene_dimer):
    coords, species, vol_ratios = benzene_dimer[0]
    alpha_0, C6, R_vdw = from_volumes(species, vol_ratios)
    ene = MBDGeom(coords, do_rpa=True, rpa_rescale_eigs=True).mbd_energy(
        alpha_0, C6, R_vdw, 0.83, variant='plain'
    )
    assert ene != approx(-0.007002398506090302, rel=1e-9)
    assert ene == approx(-0.007002398506090302, rel=1e-7)


def test_ethylcarbamate(ethylcarbamate):
    enes = [
        MBDGeom(coords, lattice, k_grid).mbd_energy_species(species, vol_ratios, 0.83)
        for coords, lattice, k_grid, species, vol_ratios in ethylcarbamate
    ]
    ene_int = enes[0] - 2 * enes[1]
    assert ene_int == approx(-0.037040868610822564, rel=1e-10)


def test_argon_crystal(argon_crystal):
    coords, lattice, k_grid, species, vol_ratios = argon_crystal
    ene = MBDGeom(coords, lattice, k_grid).mbd_energy_species(species, vol_ratios, 0.83)
    assert ene == approx(-0.0021037562496878173, rel=1e-10)


def test_argon_crystal_rpa(argon_crystal):
    coords, lattice, k_grid, species, vol_ratios = argon_crystal
    ene = MBDGeom(coords, lattice, k_grid, do_rpa=True).mbd_energy_species(
        species, vol_ratios, 0.83, variant='plain'
    )
    assert ene == approx(-0.0021036969146744147, rel=1e-10)


@pytest.mark.no_scalapack
def test_argon_crystal_modes(argon_crystal):
    coords, lattice, k_grid, species, vol_ratios = argon_crystal
    geom = MBDGeom(
        coords, lattice, custom_k_pts=[(0, 0, 0), (0.1, 0, 0)], get_spectrum=True
    )
    _, eigs, C = geom.mbd_energy_species(species, vol_ratios, 0.83)
    assert abs(C[:, :, 0].imag).sum() == approx(0)
    assert abs(C[:, :, 1].imag).sum() != approx(0)


def test_argon_crystal_gradients(argon_crystal):
    coords, lattice, k_grid, species, vol_ratios = argon_crystal
    ene, gradients, latt_gradients = MBDGeom(
        coords, lattice, k_grid
    ).mbd_energy_species(species, vol_ratios, 0.83, force=True)
    with MBDGeom(coords, lattice, k_grid) as geom:
        num_gradients = numerical_gradients(
            geom, 'mbd_energy_species', species, vol_ratios, 0.83
        )
    for i in range(len(coords)):
        assert gradients[i] == approx(num_gradients[i], rel=1e-10, abs=1e-10)
    with MBDGeom(coords, lattice, k_grid) as geom:
        num_latt_gradients = numerical_latt_gradients(
            geom, 'mbd_energy_species', species, vol_ratios, 0.83
        )
    for i in range(3):
        assert latt_gradients[i] == approx(num_latt_gradients[i], rel=1e-10, abs=1e-10)


def test_lithium(bulk_lithium):
    coords, lattice, k_grid, species, vol_ratios = bulk_lithium
    with pytest.raises(MBDFortranError):
        MBDGeom(coords, lattice, k_grid).mbd_energy_species(species, vol_ratios, 0.83)


def test_ethylcarbamate_scs(ethylcarbamate):
    enes = [
        MBDGeom(coords, lattice, k_grid).mbd_energy_species(
            species, vol_ratios, 1, a=2.56, variant='scs'
        )
        for coords, lattice, k_grid, species, vol_ratios in ethylcarbamate
    ]
    ene_int = enes[0] - 2 * enes[1]
    assert ene_int == approx(-0.03633331132194684, rel=1e-10)


def test_ethylcarbamate_ts(ethylcarbamate):
    enes = [
        MBDGeom(coords, lattice).ts_energy_species(species, vol_ratios, 0.83)
        for coords, lattice, _, species, vol_ratios in ethylcarbamate
    ]
    ene_int = enes[0] - 2 * enes[1]
    assert ene_int == approx(-0.05218213230219945, rel=1e-10)


@pytest.mark.no_scalapack
def test_mbd_coulomb(peptide_meoh):
    a = 14.4
    beta = 2.0
    enes = []
    for coords, species, vol_ratios in peptide_meoh:
        geom = MBDGeom(coords, get_spectrum=True)
        _, eigs, C = geom.mbd_energy_species(species, vol_ratios, beta=0.83)
        omega_t = np.sqrt(eigs)
        alpha_0, C6, R_vdw = from_volumes(species, vol_ratios)
        omega = 4 * C6 / (3 * alpha_0**2)
        charges = np.ones_like(alpha_0)
        masses = 1 / (alpha_0 * omega**2)
        ecoul = geom.coulomb_energy(
            charges, masses, omega_t, 'fermi', R_vdw, beta, a, C
        )
        edip = geom.dipole_energy(
            alpha_0, omega, omega_t, 'fermi,dip', R_vdw, beta, a, C
        )
        C = np.identity(len(omega_t))
        omega_non = np.repeat(omega, 3)
        ecoul_non = geom.coulomb_energy(
            charges, masses, omega_non, 'fermi', R_vdw, beta, a, C
        )
        edip_non = geom.dipole_energy(
            alpha_0, omega, omega_t, 'fermi,dip', R_vdw, beta, a, C
        )
        enes.append(ecoul - edip - (ecoul_non - edip_non))
    ene_int = enes[2] - enes[0] - enes[1]
    assert ene_int == approx(0.0002460638172163822 / 627.503, rel=1e-10)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import division, print_function

import csv
import sys
from itertools import product

import numpy as np
from numpy.polynomial.legendre import leggauss
from pkg_resources import resource_string
from scipy.special import erf, erfc

__all__ = ['mbd_energy', 'mbd_energy_species', 'screening', 'ang']

ang = 1 / 0.529177249
"""(a.u.) angstrom"""


def screening(coords, alpha_0, C6, R_vdw, beta, lattice=None, nfreq=15):
    r"""Screen atomic polarizabilities.

    :param array-like coords: (a.u.) atom coordinates in rows
    :param array-like alpha_0: (a.u.) atomic polarizabilities
    :param array-like C6: (a.u.) atomic :math:`C_6` coefficients
    :param array-like R_vdw: (a.u.) atomic vdW radii
    :param float beta: MBD damping parameter :math:`\beta`
    :param array-like lattice: (a.u.) lattice vectors in rows
    :param int nfreq: number of grid points for frequency quadrature

    Returns static polarizabilities, :math:`C_6` coefficients, and
    :math:`R_\mathrm{vdw}` coefficients (a.u.).
    """
    freq, freq_w = freq_grid(nfreq)
    omega = 4 / 3 * C6 / alpha_0**2
    alpha_dyn = [alpha_0 / (1 + (u / omega) ** 2) for u in freq]
    alpha_dyn_rsscs = []
    for a in alpha_dyn:
        sigma = (np.sqrt(2 / np.pi) * a / 3) ** (1 / 3)
        dipmat = dipole_matrix(
            coords, 'fermi,dip,gg', sigma=sigma, R_vdw=R_vdw, beta=beta, lattice=lattice
        )
        a_nlc = np.linalg.inv(np.diag(np.repeat(1 / a, 3)) + dipmat)
        a_contr = sum(np.sum(a_nlc[i::3, i::3], 1) for i in range(3)) / 3
        alpha_dyn_rsscs.append(a_contr)
    alpha_dyn_rsscs = np.stack(alpha_dyn_rsscs)
    C6_rsscs = 3 / np.pi * np.sum(freq_w[:, None] * alpha_dyn_rsscs**2, 0)
    R_vdw_rsscs = R_vdw * (alpha_dyn_rsscs[0, :] / alpha_0) ** (1 / 3)
    return alpha_dyn_rsscs[0], C6_rsscs, R_vdw_rsscs


def mbd_energy(coords, alpha_0, C6, R_vdw, beta, lattice=None, k_grid=None, nfreq=15):
    r"""Calculate an MBD energy.

    :param array-like coords: (a.u.) atom coordinates in rows
    :param array-like alpha_0: (a.u.) atomic polarizabilities
    :param array-like C6: (a.u.) atomic :math:`C_6` coefficients
    :param array-like R_vdw: (a.u.) atomic vdW radii
    :param float beta: MBD damping parameter :math:`\beta`
    :param array-like lattice: (a.u.) lattice vectors in rows
    :param array-like k_grid: number of :math:`k`-points along reciprocal axes
    :param int nfreq: number of grid points for frequency quadrature
    """
    coords, alpha_0, C6, R_vdw, lattice = map(
        _array, (coords, alpha_0, C6, R_vdw, lattice)
    )
    alpha_0_rsscs, C6_rsscs, R_vdw_rsscs = screening(
        coords, alpha_0, C6, R_vdw, beta, lattice=lattice, nfreq=15
    )
    omega_rsscs = 4 / 3 * C6_rsscs / alpha_0_rsscs**2
    pre = np.repeat(omega_rsscs * np.sqrt(alpha_0_rsscs), 3)
    if lattice is None:
        k_points = [None]
    else:
        assert k_grid is not None
        k_points = get_kpts(lattice, k_grid)
    ene = 0.0
    for k_point in k_points:
        eigs = np.linalg.eigvalsh(
            np.diag(np.repeat(omega_rsscs**2, 3))
            + np.outer(pre, pre)
            * dipole_matrix(
                coords,
                'fermi,dip',
                R_vdw=R_vdw_rsscs,
                beta=beta,
                lattice=lattice,
                k_point=k_point,
            )
        )
        ene += np.sum(np.sqrt(eigs)) / 2
    ene = ene / len(k_points) - 3 * np.sum(omega_rsscs) / 2
    return ene


def mbd_energy_species(coords, species, volume_ratios, beta, **kwargs):
    r"""Calculate an MBD energy from atom types and Hirshfed-volume ratios.

    :param array-like coords: (a.u.) atom coordinates in rows
    :param array-like species: atom types (elements)
    :param array-like volume_ratios: ratios of Hirshfeld volumes in molecule and vacuum
    :param float beta: MBD damping parameter :math:`\beta`
    :param kwargs: see :func:`mbd_energy`
    """
    alpha_0, C6, R_vdw = from_volumes(species, volume_ratios)
    return mbd_energy(coords, alpha_0, C6, R_vdw, beta, **kwargs)


def dipole_matrix(
    coords, damping, beta=0.0, lattice=None, k_point=None, R_vdw=None, sigma=None, a=6.0
):
    if lattice is not None:
        volume = max(np.abs(np.product(np.linalg.eigvals(lattice))), 0.2)
        ewald_alpha = 2.5 / volume ** (1 / 3)
        real_space_cutoff = 6 / ewald_alpha
        range_cell = supercell_circum(lattice, real_space_cutoff)
    else:
        range_cell = (0, 0, 0)
    do_ewald = lattice is not None and damping in {'fermi,dip'}
    n = len(coords)
    dtype = float if k_point is None else complex
    dipmat = np.zeros((n, n, 3, 3), dtype=dtype)
    if R_vdw is not None:
        S_vdw = beta * (R_vdw[:, None] + R_vdw[None, :])
    if sigma is not None:
        sigma_ij = np.sqrt(sigma[:, None] ** 2 + sigma[None, :] ** 2)
    for idx_cell in product(*(range(-i, i + 1) for i in range_cell)):
        R_cell = lattice.T.dot(idx_cell) if lattice is not None else np.zeros(3)
        Rs = coords[:, None, :] - coords[None, :, :] + R_cell
        dists = np.sqrt(np.sum(Rs**2, -1))
        if damping == 'fermi,dip':
            T = damping_fermi(dists, S_vdw, a)[:, :, None, None] * T_bare(Rs)
        elif damping == 'fermi,dip,gg':
            T = (1 - damping_fermi(dists, S_vdw, a)[:, :, None, None]) * T_erf_coulomb(
                Rs, sigma_ij
            )
        else:
            raise ValueError(f'Unsupported damping: {damping}')
        if do_ewald:
            T += T_erfc(Rs, ewald_alpha) - T_bare(Rs)
        if k_point is not None:
            k_pref = np.exp(-1j * np.sum(k_point * Rs, -1))
            T = k_pref[:, :, None, None] * T
        dipmat += T
    if do_ewald:
        dipmat += dipole_matrix_ewald(coords, lattice, ewald_alpha, k_point)
    n = len(coords)
    return np.reshape(np.transpose(dipmat, (0, 2, 1, 3)), (3 * n, 3 * n))


def dipole_matrix_ewald(coords, lattice, alpha, k_point=None):
    Rs = coords[:, None, :] - coords[None, :, :]
    rlattice = 2 * np.pi * np.linalg.inv(lattice.T)
    volume = abs(np.product(np.linalg.eigvals(lattice)))
    rec_space_cutoff = 10 * alpha
    range_G_vector = supercell_circum(rlattice, rec_space_cutoff)
    dtype = float if k_point is None else complex
    dipmat = np.zeros((len(Rs), len(Rs), 3, 3), dtype=dtype)
    fourier_factor = (
        (lambda x: np.cos(x)) if k_point is None else (lambda x: np.exp(1j * x))
    )
    for idx_gvec in product(*(range(-i, i + 1) for i in range_G_vector)):
        if idx_gvec == (0, 0, 0):
            continue
        gvec = rlattice.T.dot(idx_gvec)
        k_total = k_point + gvec if k_point is not None else gvec
        k_sq = sum(k_total**2)
        if np.sqrt(k_sq) > rec_space_cutoff:
            continue
        k_prefactor = (
            4
            * np.pi
            / volume
            * np.exp(-k_sq / (4 * alpha**2))
            / k_sq
            * np.outer(k_total, k_total)
        )
        dipmat += k_prefactor * fourier_factor(np.sum(gvec * Rs, -1))[:, :, None, None]
    dipmat += -np.eye(len(Rs))[:, :, None, None] * np.diag(
        np.repeat(4 * alpha**3 / (3 * np.sqrt(np.pi)), 3)
    )
    k_sq = np.sum(k_point**2) if k_point is not None else 0
    if np.sqrt(k_sq) > 1e-15:
        dipmat += (
            4
            * np.pi
            / volume
            * np.exp(-k_sq / (4 * alpha**2))
            / k_sq
            * np.outer(k_point, k_point)
        )
    else:
        dipmat += np.diag(np.repeat(4 * np.pi / (3 * volume), 3))
    return dipmat


def supercell_circum(latt, radius):
    rlatt = 2 * np.pi * np.linalg.inv(latt.T)
    layer_sep = np.sum(latt * rlatt / np.sqrt(np.sum(rlatt**2, 1))[None, :], 0)
    return np.ceil(radius / layer_sep + 0.5).astype(int)


def damping_fermi(R, S_vdw, d):
    return 1 / (1 + np.exp(-d * (R / S_vdw - 1)))


def T_bare(R):
    R_2 = np.sum(R**2, -1)
    R_5 = np.where(R_2 > 0, np.sqrt(R_2) ** 5, np.inf)
    return (
        -3 * R[:, :, :, None] * R[:, :, None, :]
        + R_2[:, :, None, None] * np.eye(3)[None, None, :, :]
    ) / R_5[:, :, None, None]


def T_erf_coulomb(R, sigma):
    bare = T_bare(R)
    R_1 = np.sqrt(np.sum(R**2, -1))
    R_5 = np.where(R_1 > 0, R_1**5, np.inf)
    RR_R5 = R[:, :, :, None] * R[:, :, None, :] / R_5[:, :, None, None]
    zeta = R_1 / sigma
    theta = 2 * zeta / np.sqrt(np.pi) * np.exp(-(zeta**2))
    erf_theta = erf(zeta) - theta
    return (
        erf_theta[:, :, None, None] * bare
        + (2 * (zeta**2) * theta)[:, :, None, None] * RR_R5
    )


def T_erfc(R, a):
    R_2 = np.sum(R**2, -1)
    R_1 = np.sqrt(R_2)
    R_3 = np.where(R_1 > 0, R_1**3, np.inf)
    R_5 = np.where(R_1 > 0, R_1**5, np.inf)
    B = (
        erfc(a * R_1) + (2 * a * R_1 / np.sqrt(np.pi)) * np.exp(-((a * R_1) ** 2))
    ) / R_3
    C = (
        3 * erfc(a * R_1)
        + (2 * a * R_1 / np.sqrt(np.pi))
        * (3 + 2 * (a * R_1) ** 2)
        * np.exp(-((a * R_1) ** 2))
    ) / R_5
    return -C[:, :, None, None] * R[:, :, :, None] * R[:, :, None, :] + B[
        :, :, None, None
    ] * np.eye(3)


def from_volumes(species, volumes, kind='TS'):
    if kind == 'TS':
        alpha_0, C6, R_vdw = (
            np.array([vdw_params[sp][param] for sp in species])
            for param in 'alpha_0(TS) C6(TS) R_vdw(TS)'.split()
        )
    elif kind == 'BG':
        alpha_0, C6, R_vdw = (
            np.array([vdw_params[sp][param] for sp in species])
            for param in 'alpha_0(BG) C6(BG) R_vdw(TS)'.split()
        )
    elif kind == 'TSsurf':
        alpha_0, C6, R_vdw = (
            np.array(
                [
                    vdw_params[sp][param]
                    or vdw_params[sp][param.replace('TSsurf', 'TS')]
                    for sp in species
                ]
            )
            for param in 'alpha_0(TSsurf) C6(TSsurf) R_vdw(TSsurf)'.split()
        )
    else:
        raise ValueError(f'Unkonwn vdW parameter kind: {kind}')
    volumes = np.array(volumes)
    alpha_0 *= volumes
    C6 *= volumes**2
    R_vdw *= volumes ** (1 / 3)
    return alpha_0, C6, R_vdw


def get_kpts(lattice, k_grid, shift=0.5):
    k_grid, lattice = map(np.array, (k_grid, lattice))
    k_idxs = (np.array(list(product(*map(range, k_grid)))) + shift) / k_grid
    k_idxs = np.where(k_idxs > 0.5, k_idxs - 1, k_idxs)
    rlattice = 2 * np.pi * np.linalg.inv(lattice.T)
    k_points = k_idxs.dot(rlattice)
    return k_points


def freq_grid(n, L=0.6):
    x, w = leggauss(n)
    w = 2 * L / (1 - x) ** 2 * w
    x = L * (1 + x) / (1 - x)
    return np.hstack(([0], x[::-1])), np.hstack(([0], w[::-1]))


def _array(obj, *args, **kwargs):
    if obj is not None:
        kwargs.setdefault('dtype', float)
        return np.array(obj, *args, **kwargs)


def _get_vdw_params():
    csv_lines = resource_string(__name__, 'vdw-params.csv').split(b'\n')
    if sys.version_info[0] > 2:
        csv_lines = [l.decode() for l in csv_lines]
    reader = csv.DictReader(csv_lines, quoting=csv.QUOTE_NONNUMERIC)
    vdw_params = {row.pop('symbol'): row for row in reader}
    return vdw_params


vdw_params = _get_vdw_params()



#!/usr/bin/env python3
import re

tests = []
with open('tests/mbd_grad_test_cases.F90') as f:
    for l in f:
        m = re.match(r' *subroutine *test_(\w+) *\(', l)
        if m:
            tests.append(f'grad/{m.group(1)}')
with open('tests/mbd_api_tests.F90') as f:
    for l in f:
        m = re.match(r' *case *\( *["\'](\w+)["\'] *\)', l)
        if m:
            tests.append(f'api/{m.group(1)}')
print(';'.join(tests))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import division, print_function

import re
from functools import wraps

import numpy as np

from .__init__ import __version__ as PYMBD_VERSION
from .pymbd import _array, from_volumes

try:
    from ._libmbd import ffi as _ffi, lib as _lib
except ImportError:
    raise Exception('pyMBD C extension unimportable, cannot use Fortran') from None

__all__ = ['MBDGeom', 'with_mpi', 'with_scalapack']

with_mpi = _lib.cmbd_with_mpi
"""Whether libMBD was compiled with MPI"""

with_scalapack = _lib.cmbd_with_scalapack
"""Whether libMBD was compiled with Scalapack"""

LIBMBD_VERSION = (
    _lib.cmbd_version_major,
    _lib.cmbd_version_minor,
    _lib.cmbd_version_patch,
    _ffi.string(_lib.cmbd_version_suffix).strip().decode(),
)

# do not test versions when running autodoc
if PYMBD_VERSION and isinstance(LIBMBD_VERSION[0], int):
    assert PYMBD_VERSION[0] == LIBMBD_VERSION[0]
    if PYMBD_VERSION[0] == 0:
        assert PYMBD_VERSION[1] == LIBMBD_VERSION[1]
    else:
        assert PYMBD_VERSION[1] <= LIBMBD_VERSION[1]
    if len(PYMBD_VERSION) == 4:
        git_commit = re.split('[+-]', PYMBD_VERSION[3])[1]
        assert LIBMBD_VERSION[3].endswith(git_commit)


class MBDFortranError(Exception):
    def __init__(self, code, origin, msg):
        super(MBDFortranError, self).__init__(msg)
        self.code = code
        self.origin = origin


def _auto_context(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self._geom_f:
            return method(self, *args, **kwargs)
        with self:
            return method(self, *args, **kwargs)

    return wrapper


class MBDGeom(object):
    """Represents an initialized libMBD `geom_t <../type/geom_t.html>`_ object.

    :param array-like coords: (a.u.) atomic coordinates as rows
    :param array-like lattice: (a.u.) lattice vectors as rows
    :param array-like k_grid: number of :math:`k`-points per reciprocal lattice vector
    :param array-like custom_k_pts: (a.u.) custom :math:`k`-points as rows
    :param int n_freq: number of quadrature points for frequency integration
    :param bool do_rpa: whether to calculate MBD energy via frequency integration
    :param bool get_spectrum: whether to return eigenvalues and eigenvectors
    :param bool get_rpa_orders: whether to return RPA order decomposition
    :param bool rpa_rescale_eigs: whether to rescale RPA eigenvalues
    """

    def __init__(
        self,
        coords,
        lattice=None,
        k_grid=None,
        custom_k_pts=None,
        n_freq=None,
        do_rpa=False,
        get_spectrum=False,
        get_rpa_orders=False,
        rpa_rescale_eigs=False,
        max_atoms_per_block=None,
        ewald_cutoff_scaling=(1.0, 1.0),
    ):
        self._geom_f = None
        self._coords, self._lattice = map(_array, (coords, lattice))
        self._k_grid = _array(k_grid, dtype='i4')
        self._custom_k_pts = _array(custom_k_pts)
        self._n_freq = n_freq
        self._do_rpa = do_rpa
        self._get_spectrum = get_spectrum
        self._get_rpa_orders = get_rpa_orders
        self._rpa_rescale_eigs = rpa_rescale_eigs
        self._max_atoms_per_block = max_atoms_per_block
        self._ewald_cutoff_scaling = ewald_cutoff_scaling

    def __len__(self):
        return len(self._coords)

    def __enter__(self):
        self._geom_f = _lib.cmbd_init_geom(
            len(self),
            _cast('double*', self._coords),
            _cast('double*', self._lattice),
            _cast('int*', self._k_grid),
            len(self._custom_k_pts) if self._custom_k_pts is not None else 0,
            _cast('double*', self._custom_k_pts),
            self._n_freq or 0,
            self._do_rpa,
            self._get_spectrum,
            self._get_rpa_orders,
            self._rpa_rescale_eigs,
            self._max_atoms_per_block or 0,
            self._ewald_cutoff_scaling,
        )
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        _lib.cmbd_destroy_geom(self._geom_f)
        self._geom_f = None

    def _check_exc(self):
        code = _array(0, dtype=int)
        origin = _ffi.new('char[50]')
        msg = _ffi.new('char[150]')
        _lib.cmbd_get_exception(self._geom_f, _cast('int*', code), origin, msg)
        if code != 0:
            raise MBDFortranError(
                int(code), _ffi.string(origin).decode(), _ffi.string(msg).decode()
            )

    @property
    def coords(self):
        """(a.u.) Atom coordinates in rows."""
        return self._coords.copy()

    @coords.setter
    def coords(self, coords):
        _lib.cmbd_update_coords(self._geom_f, _cast('double*', _array(coords)))

    @property
    def lattice(self):
        """(a.u.) Lattice vectors in rows."""
        return self._lattice.copy()

    @lattice.setter
    def lattice(self, lattice):
        _lib.cmbd_update_lattice(self._geom_f, _cast('double*', _array(lattice)))

    def has_lattice(self):
        """Whether structure is a crystal."""
        return self._lattice is not None

    def print_timing(self):
        """Print timing from libMBD."""
        _lib.cmbd_print_timing(self._geom_f)

    @_auto_context
    def ts_energy(self, alpha_0, C6, R_vdw, sR, d=20.0, damping='fermi', force=False):
        """Calculate a TS energy.

        :param array-like alpha_0: (a.u.) atomic polarizabilities
        :param array-like C6: (a.u.) atomic :math:`C_6` coefficients
        :param array-like R_vdw: (a.u.) atomic vdW radii
        :param float sR: TS damping parameter :math:`s_R`
        :param float d: TS damping parameter :math:`d`
        :param damping str: type of damping
        :param force bool: if True, calculate energy gradients
        """
        alpha_0, C6, R_vdw = map(_array, (alpha_0, C6, R_vdw))
        n_atoms = len(self)
        damping_f = _lib.cmbd_init_damping(
            len(self), damping.encode(), _cast('double*', R_vdw), _ffi.NULL, sR, d
        )
        res_f = _lib.cmbd_ts_energy(
            self._geom_f,
            _cast('double*', alpha_0),
            _cast('double*', C6),
            damping_f,
            force,
        )
        _lib.cmbd_destroy_damping(damping_f)
        self._check_exc()
        ene = np.empty(1)  # for some reason np.array(0) doesn't work
        gradients, lattice_gradients = 2 * [None]
        if force:
            gradients = np.zeros((n_atoms, 3))
            if self.has_lattice():
                lattice_gradients = np.zeros((3, 3))
        results = (ene, gradients, lattice_gradients, *(7 * [None]))
        _lib.cmbd_get_results(res_f, *(_cast('double*', x) for x in results))
        _lib.cmbd_destroy_result(res_f)
        ene = ene.item()
        if force:
            ene = (ene, gradients)
            if self.has_lattice():
                ene += (lattice_gradients,)
        return ene

    @_auto_context
    def mbd_energy(  # noqa: C901
        self,
        alpha_0,
        C6,
        R_vdw=None,
        beta=0.0,
        a=6.0,
        sigma=None,
        damping='fermi,dip',
        variant='rsscs',
        force=False,
        intermediates=False,
    ):
        r"""Calculate an MBD energy.

        :param array-like alpha_0: (a.u.) atomic polarizabilities
        :param array-like C6: (a.u.) atomic :math:`C_6` coefficients
        :param array-like R_vdw: (a.u.) atomic vdW radii
        :param array-like sigma: (a.u.) oscillator widths
        :param float beta: MBD damping parameter :math:`\beta`
        :param float a: MBD damping parameter :math:`a`
        :param damping str: type of damping
        :param variant str: one of 'plain', 'scs', 'rsscs'
        :param force bool: if True, calculate energy gradients
        """
        alpha_0, C6, R_vdw, sigma = map(_array, (alpha_0, C6, R_vdw, sigma))
        n_atoms = len(self)
        damping_f = _lib.cmbd_init_damping(
            n_atoms,
            damping.encode(),
            _cast('double*', R_vdw),
            _cast('double*', sigma),
            beta,
            a,
        )
        args = (
            self._geom_f,
            _cast('double*', alpha_0),
            _cast('double*', C6),
            damping_f,
            force,
        )
        if variant == 'plain':
            res_f = _lib.cmbd_mbd_energy(*args)
        else:
            args = args[:1] + (variant.encode(),) + args[1:]
            res_f = _lib.cmbd_mbd_scs_energy(*args)
        _lib.cmbd_destroy_damping(damping_f)
        self._check_exc()
        ene = np.empty(1)  # for some reason np.array(0) doesn't work
        gradients, lattice_gradients = 2 * [None]
        alpha_0_scs, C6_scs = 2 * [None]
        eigs, modes, rpa_orders, eigs_k, modes_k = 5 * [None]
        if force:
            gradients = np.zeros((n_atoms, 3))
            if self.has_lattice():
                lattice_gradients = np.zeros((3, 3))
        if intermediates:
            alpha_0_scs, C6_scs = np.zeros(n_atoms), np.zeros(n_atoms)
        if self._get_spectrum:
            if self.has_lattice():
                n_kpts = (
                    len(self._custom_k_pts)
                    if self._custom_k_pts is not None
                    else self._k_grid.prod()
                )
                eigs_k = np.zeros((3 * n_atoms, n_kpts), order='F')
                modes_k = np.zeros(
                    (3 * n_atoms, 3 * n_atoms, n_kpts), dtype=complex, order='F'
                )
            else:
                eigs = np.zeros(3 * n_atoms)
                modes = np.zeros((3 * n_atoms, 3 * n_atoms), order='F')
        elif self._get_rpa_orders:
            rpa_orders = np.zeros(10)
        results = (
            ene,
            gradients,
            lattice_gradients,
            eigs,
            modes,
            rpa_orders,
            eigs_k,
            modes_k,
            alpha_0_scs,
            C6_scs,
        )
        _lib.cmbd_get_results(res_f, *(_cast('double*', x) for x in results))
        _lib.cmbd_destroy_result(res_f)
        ene = ene.item()
        if self.has_lattice():
            eigs, modes = eigs_k, modes_k
        if self._get_spectrum:
            ene = ene, eigs, modes
        elif self._get_rpa_orders:
            ene = ene, rpa_orders
        if force or intermediates:
            ene = (ene,)
        if force:
            ene += (gradients,)
            if self.has_lattice():
                ene += (lattice_gradients,)
        if intermediates:
            ene += (alpha_0_scs, C6_scs)
        return ene

    @_auto_context
    def dipole_matrix(
        self, damping, beta=0.0, k_point=None, R_vdw=None, sigma=None, a=6.0
    ):  # noqa: D102
        R_vdw, sigma, k_point = map(_array, (R_vdw, sigma, k_point))
        n_atoms = len(self)
        damping_f = _lib.cmbd_init_damping(
            n_atoms,
            damping.encode(),
            _cast('double*', R_vdw),
            _cast('double*', sigma),
            beta,
            a,
        )
        dipmat = np.empty(
            (3 * n_atoms, 3 * n_atoms), dtype=float if k_point is None else complex
        )
        _lib.cmbd_dipole_matrix(
            self._geom_f, damping_f, _cast('double*', k_point), _cast('double*', dipmat)
        )
        _lib.cmbd_destroy_damping(damping_f)
        self._check_exc()
        return dipmat

    def mbd_energy_species(self, species, volume_ratios, beta, **kwargs):
        r"""Calculate an MBD energy from atom types and Hirshfed-volume ratios.

        :param array-like species: atom types (elements)
        :param array-like volume_ratios: ratios of Hirshfeld volumes in
            molecule and vacuum
        :param float beta: MBD damping parameter :math:`\beta`
        :param kwargs: see :meth:`mbd_energy`
        """
        alpha_0, C6, R_vdw = from_volumes(species, volume_ratios)
        return self.mbd_energy(alpha_0, C6, R_vdw, beta, **kwargs)

    def ts_energy_species(self, species, volume_ratios, beta, **kwargs):
        """Calculate a TS energy from atom types and Hirshfed-volume ratios.

        :param array-like species: atom types (elements)
        :param array-like volume_ratios: ratios of Hirshfeld volumes in
            molecule and vacuum
        :param float sR: TS damping parameter :math:`s_R`
        :param kwargs: see :meth:`ts_energy`
        """
        alpha_0, C6, R_vdw = from_volumes(species, volume_ratios)
        return self.ts_energy(alpha_0, C6, R_vdw, beta, **kwargs)

    @_auto_context
    def dipole_energy(self, a0, w, w_t, version, r_vdw, beta, a, C):  # noqa: D102
        n_atoms = len(self)
        res = _lib.cmbd_dipole_energy(
            self._geom_f,
            n_atoms,
            _cast('double*', a0),
            _cast('double*', w),
            _cast('double*', w_t),
            version.encode(),
            _cast('double*', r_vdw),
            beta,
            a,
            _cast('double*', C),
        )
        self._check_exc()
        return res

    @_auto_context
    def coulomb_energy(self, q, m, w_t, version, r_vdw, beta, a, C):  # noqa: D102
        n_atoms = len(self)
        res = _lib.cmbd_coulomb_energy(
            self._geom_f,
            n_atoms,
            _cast('double*', q),
            _cast('double*', m),
            _cast('double*', w_t),
            version.encode(),
            _cast('double*', r_vdw),
            beta,
            a,
            _cast('double*', C),
        )
        self._check_exc()
        return res

    @_auto_context
    def nonint_density(self, pts, q, m, w):  # noqa: D102
        n_pts = len(pts)
        n_atoms = len(self)
        rho = np.empty(n_pts)
        _lib.cmbd_nonint_density(
            self._geom_f,
            n_atoms,
            n_pts,
            _cast('double*', pts),
            _cast('double*', q),
            _cast('double*', m),
            _cast('double*', w),
            _cast('double*', rho),
        )
        self._check_exc()
        return rho

    @_auto_context
    def int_density(self, pts, q, m, w_t, modes):  # noqa: D102
        n_pts = len(pts)
        n_atoms = len(self)
        rho = np.empty(n_pts)
        _lib.cmbd_int_density(
            self._geom_f,
            n_atoms,
            n_pts,
            _cast('double*', pts),
            _cast('double*', q),
            _cast('double*', m),
            _cast('double*', w_t),
            _cast('double*', modes),
            _cast('double*', rho),
        )
        self._check_exc()
        return rho


def _ndarray(ptr, shape=None, dtype='float'):
    buffer_size = (np.prod(shape) if shape else 1) * np.dtype(dtype).itemsize
    return np.ndarray(buffer=_ffi.buffer(ptr, buffer_size), shape=shape, dtype=dtype)


def _cast(ctype, array):
    return _ffi.NULL if array is None else _ffi.cast(ctype, array.ctypes.data)



#!/usr/bin/env python3
import re
import subprocess
from string import Template

END = 'END'
s = subprocess.run(
    'git tag -l "*.*.*"'
    ' --format "%(refname:strip=2),%(taggerdate:short),'
    f'%(contents:subject)\n\n%(contents:body){END}"'
    ' --sort=-creatordate',
    shell=True,
    capture_output=True,
).stdout.decode()
data = [
    (v, d, m.strip() if m[0] == '#' else None)
    for v, d, m in re.findall(rf'(\d+\.\d+\.\d+),([\d-]+),(.*?){END}', s, re.DOTALL)
]
with open('devtools/CHANGELOG.md.in') as f:
    t = Template(f.read())
changes = [f'## [{v}] - {d}\n\n{m}' for v, d, m in data if m]
link = '[{0}]: https://github.com/libmbd/libmbd/{1}'
vs = [x[0] for x in data]
links = [
    *(
        link.format(
            vs[i - 1] if i else 'unreleased',
            f'compare/{vs[i]}...{vs[i-1] if i else "HEAD"}',
        )
        for i in range(len(vs))
    ),
    link.format(vs[-1], f'releases/tag/{vs[-1]}'),
]
with open('wiki/Changelog.md', 'w') as f:
    f.write(t.substitute({'changes': '\n\n'.join(changes), 'links': '\n'.join(links)}))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import sys

from pymbd import ang
from pymbd.fortran import MBDGeom, with_mpi

if with_mpi:
    from mpi4py import MPI

    rank = MPI.COMM_WORLD.Get_rank()
else:
    rank = 0

ene_expected = -0.0002462647623815428
ene = MBDGeom([(0, 0, 0), (0, 0, 4 * ang)]).mbd_energy_species(
    ['Ar', 'Ar'], [1, 1], 0.83
)
if rank == 0:
    print(f'Expected energy:   {ene_expected}')
    print(f'Calculated energy: {ene}')
if ene - ene_expected > 1e-10:
    sys.exit(1)



import datetime
import os
import subprocess
import sys

import toml

sys.path.insert(0, os.path.abspath('../src'))


with open('../pyproject.toml') as f:
    metadata = toml.load(f)['tool']['poetry']

project = 'libMBD'
release = version = (
    subprocess.run(['poetry', 'version'], capture_output=True, cwd='..')
    .stdout.decode()
    .split()[1]
)
author = ' '.join(metadata['authors'][0].split()[:-1])
description = metadata['description']

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
]
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}
source_suffix = '.rst'
master_doc = 'index'
copyright = f'2018-{datetime.date.today().year}, {author}'
language = None
exclude_patterns = ['build', '.DS_Store']
pygments_style = 'sphinx'
todo_include_todos = True
html_theme = 'alabaster'
html_theme_options = {
    'description': description,
    'github_button': True,
    'github_user': 'libmbd',
    'github_repo': 'libmbd',
    'badge_branch': 'master',
    'codecov_button': True,
}
html_sidebars = {
    '**': ['about.html', 'navigation.html', 'relations.html', 'searchbox.html']
}

autodoc_default_options = {'members': True}
autodoc_inherit_docstrings = False
autodoc_mock_imports = [
    'numpy',
    'scipy',
    'pymbd._libmbd',
    'mpi4py',
]


def skip_namedtuples(app, what, name, obj, skip, options):
    if hasattr(obj, '_source'):
        return True


def setup(app):
    app.connect('autodoc-skip-member', skip_namedtuples)



#!/usr/bin/env python3
import sys
from argparse import ArgumentParser
from functools import partial

import numpy as np

from pymbd import __version__ as _version, ang
from pymbd.fortran import LIBMBD_VERSION, MBDGeom, with_mpi

__all__ = ()

# fmt: off
unit_cell = (
    np.array([
        (4.083, 5.700, 2.856), (0.568, 0.095, 4.217), (0.470, 4.774, 3.551),
        (4.181, 1.022, 3.522), (5.572, 5.587, 1.892), (-0.920, 0.209, 5.181),
        (3.663, 3.255, 2.585), (0.988, 2.541, 4.488), (3.834, 4.011, 0.979),
        (0.816, 1.785, 6.094), (2.223, 1.314, 1.108), (2.428, 4.481, 5.965),
        (1.177, 0.092, 0.406), (3.474, 5.703, 6.667), (4.911, 5.036, 2.573),
        (-0.260, 0.759, 4.500), (4.358, 3.787, 1.918), (0.293, 2.009, 5.155),
        (0.205, 1.729, 1.101), (4.446, 4.067, 5.972), (1.285, 0.947, 0.957),
        (3.366, 4.848, 6.116), (0.485, 2.901, 1.709), (4.165, 2.895, 5.364),
        (4.066, 1.426, 0.670), (0.585, 4.369, 6.403)]) * ang,
    np.array([
        (5.008, 0.018, -0.070),
        (1.630, 6.759, 0.064),
        (-1.987, -0.981, 7.079)]) * ang,
    list('HHHHHHHHHHHHHHCCCCCCNNOOOO'),
    [0.703, 0.703, 0.726, 0.726, 0.731, 0.731,
     0.727, 0.727, 0.754, 0.754, 0.750, 0.750,
     0.755, 0.755, 0.809, 0.809, 0.827, 0.827,
     0.834, 0.834, 0.840, 0.840, 0.886, 0.886,
     0.892, 0.892])
# fmt: on

RANK = None


def parse(output):
    blocks = [
        [l.split() for l in block.strip().split('\n')]
        for block in output.split('--------------')[1:-1]
    ]
    n_atoms = int(blocks[0][0][-1])
    version_pymbd = blocks[0][1][-1]
    version_libmbd = blocks[0][2][-1]
    block_size, grid = None, None
    block = iter(blocks[1])
    for words in block:
        if words[0] == 'id':
            break
        elif words[0] == 'BLACS':
            if words[1] == 'block':
                block_size = int(words[-1])
            else:
                grid = [int(words[-3]), int(words[-1])]
    timing = [
        {
            'id': int(words[0]),
            'level': int(words[1]),
            'label': ' '.join(words[2:-2]),
            'count': int(words[-2]),
            'time': float(words[-1]),
        }
        for words in block
    ]
    energy = float(blocks[2][0][-1])
    return {
        'version_pymbd': version_pymbd,
        'version_libmbd': version_libmbd,
        'n_atoms': n_atoms,
        'timing': timing,
        'energy': energy,
        'block_size': block_size,
        'grid': grid,
    }


def make_supercell(coords, lattice, species, vol_ratios, sc):
    sc = np.array(sc)
    n_uc = np.product(sc)
    c = np.stack(
        np.meshgrid(range(sc[0]), range(sc[1]), range(sc[2])), axis=-1
    ).reshape(-1, 3)
    coords = ((c @ lattice)[:, None, :] + coords).reshape(-1, 3)
    lattice = lattice * sc[:, None]
    species = n_uc * species
    vol_ratios = n_uc * vol_ratios
    return coords, lattice, species, vol_ratios


def _print(*args):
    if not RANK:
        print(*args)


def run(supercell, k_grid, finite, force, method, early_return, repeat):
    if with_mpi:
        from mpi4py import MPI

        global RANK
        RANK = MPI.COMM_WORLD.Get_rank()
    _print('--------------')
    coords, lattice, species, vol_ratios = make_supercell(*unit_cell, supercell)
    _print('number of atoms:', len(coords))
    _print('pyMBD version:', '.'.join(map(str, _version)))
    _print('libMBD version:', '{}.{}.{}-{}'.format(*LIBMBD_VERSION))
    _print('--------------')
    if early_return:
        return
    geom = MBDGeom(coords) if finite else MBDGeom(coords, lattice, k_grid)
    if method == 'mbd@rsscs':
        func = geom.mbd_energy_species
    elif method == 'mbd':
        func = partial(geom.mbd_energy_species, variant='plain')
    elif method == 'ts':
        func = geom.ts_energy_species
    with geom:
        for _ in range(repeat):
            res = func(species, vol_ratios, 0.83, force=force)
        geom.print_timing()
    if force:
        ene, grad, *_ = res
    else:
        ene = res
    ene = ene / len(coords)
    _print('--------------')
    _print('energy:', ene)
    if force:
        _print('grad[0]:', grad[0])
    _print('--------------')


def main(args):
    parser = ArgumentParser()
    parser.add_argument(
        '--parse',
        action='store_true',
        help='parse stdin and output JSON',
    )
    parser.add_argument(
        '--supercell',
        nargs=3,
        type=int,
        default=[1, 1, 1],
        metavar='N',
        help='supercell definition',
    )
    parser.add_argument(
        '--k-grid',
        nargs=3,
        type=int,
        default=[1, 1, 1],
        metavar='N',
        help='k-grid definition',
    )
    parser.add_argument(
        '--finite',
        action='store_true',
        help='run a finite system',
    )
    parser.add_argument(
        '--no-force',
        action='store_false',
        dest='force',
        help='switch off calculation of gradients',
    )
    parser.add_argument(
        '--method',
        default='mbd@rsscs',
        choices=['mbd@rsscs', 'mbd', 'ts'],
        help='method to run',
    )
    parser.add_argument(
        '--early-return',
        action='store_true',
        help='return before doing any work',
    )
    parser.add_argument(
        '--repeat',
        type=int,
        default=1,
        help='run benchmark repeatedly',
    )
    args = vars(parser.parse_args(args))
    if args.pop('parse', False):
        from pprint import pprint

        pprint(parse(sys.stdin.read()))
    else:
        run(**args)


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3
import csv
from importlib.resources import open_text

with open_text('pymbd', 'vdw-params.csv') as f:
    reader = csv.DictReader(f, quoting=csv.QUOTE_NONNUMERIC)
    rows = list(reader)
rows = [row for row in rows if row['symbol'][-1] not in '+-']
n = len(rows)
print(f'real(dp), parameter :: ts_vdw_params(3, {n}) = reshape([ &')
for i, row in enumerate(rows):
    alpha_0 = row['alpha_0(TS)'] or 0
    C6 = row['C6(TS)'] or 0
    R_vdw = row['R_vdw(TS)'] or 0
    print(
        f'    {alpha_0}d0, {C6}d0, {R_vdw}d0{"," if i < n-1 else ""} &'
        f'  ! {row["symbol"]}'
    )
print(f'], [3, {n}])')
print()
print(f'real(dp), parameter :: tssurf_vdw_params(3, {n}) = reshape([ &')
for i, row in enumerate(rows):
    alpha_0 = row['alpha_0(TSsurf)'] or row['alpha_0(TS)'] or 0
    C6 = row['C6(TSsurf)'] or row['C6(TS)'] or 0
    R_vdw = row['R_vdw(TSsurf)'] or row['R_vdw(TS)'] or 0
    print(
        f'    {alpha_0}d0, {C6}d0, {R_vdw}d0{"," if i < n-1 else ""} &'
        f'  ! {row["symbol"]}'
    )
print(f'], [3, {n}])')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import division, print_function

from math import pi

import numpy as np
import tensorflow as tf

from .pymbd import freq_grid

__all__ = ()

pi = tf.constant(pi, tf.float64)
ang = 1 / 0.529177249


class MBDEvaluator(object):
    def __init__(self, gradients=False, **kwargs):
        self._inputs = coords, alpha_0, C6, R_vdw, beta = [
            tf.placeholder(tf.float64, shape=shape, name=name)
            for shape, name in [
                ((None, 3), 'coords'),
                ((None,), 'alpha_0'),
                ((None,), 'C6'),
                ((None,), 'R_vdw'),
                ((), 'beta'),
            ]
        ]
        self._output = mbd_energy(*self._inputs, **kwargs)
        if gradients:
            self._init_gradients()
        else:
            self._gradients = None

    def _init_gradients(self):
        self._gradients = tf.gradients(self._output, [self._inputs[0]])[0]

    def __call__(self, coords, alpha_0, C6, R_vdw, beta=0.83, gradients=None):
        inputs = dict(zip(self._inputs, [coords, alpha_0, C6, R_vdw, beta]))
        outputs = self._output
        if gradients or gradients is None and self._gradients is not None:
            if self._gradients is None:
                self._init_gradients()
            outputs = outputs, self._gradients
        return tf.get_default_session().run(outputs, inputs)


def mbd_energy(coords, alpha_0, C6, R_vdw, beta, nfreq=15):
    freq, freq_w = freq_grid(nfreq)
    omega = 4 / 3 * C6 / alpha_0**2
    alpha_dyn = [alpha_0 / (1 + (u / omega) ** 2) for u in freq]
    alpha_dyn_rsscs = []
    for a in alpha_dyn:
        sigma = (tf.sqrt(2 / pi) * a / 3) ** (1 / 3)
        dipmat = dipole_matrix(
            coords, 'fermi,dip,gg', sigma=sigma, R_vdw=R_vdw, beta=beta
        )
        a_nlc = tf.linalg.inv(tf.diag(_repeat(1 / a, 3)) + dipmat)
        a_contr = sum(tf.reduce_sum(a_nlc[i::3, i::3], 1) for i in range(3)) / 3
        alpha_dyn_rsscs.append(a_contr)
    alpha_dyn_rsscs = tf.stack(alpha_dyn_rsscs)
    C6_rsscs = 3 / pi * tf.reduce_sum(freq_w[:, None] * alpha_dyn_rsscs**2, 0)
    R_vdw_rsscs = R_vdw * (alpha_dyn_rsscs[0, :] / alpha_0) ** (1 / 3)
    omega_rsscs = 4 / 3 * C6_rsscs / alpha_dyn_rsscs[0, :] ** 2
    dipmat = dipole_matrix(coords, 'fermi,dip', R_vdw=R_vdw_rsscs, beta=beta)
    pre = _repeat(omega_rsscs * tf.sqrt(alpha_dyn_rsscs[0, :]), 3)
    eigs = tf.linalg.eigvalsh(
        tf.diag(_repeat(omega_rsscs**2, 3)) + pre[:, None] * pre[None, :] * dipmat
    )
    ene = tf.reduce_sum(tf.sqrt(eigs)) / 2 - 3 * tf.reduce_sum(omega_rsscs) / 2
    return ene


def dipole_matrix(coords, damping, beta=0.0, R_vdw=None, sigma=None, a=6.0):
    Rs = coords[:, None, :] - coords[None, :, :]
    if R_vdw is not None:
        S_vdw = beta * (R_vdw[:, None] + R_vdw[None, :])
        # 1e10 rather than inf is necessary to avoid nan gradients
        dists = tf.sqrt(_set_diag(tf.reduce_sum(Rs**2, -1), 1e10))
    if sigma is not None:
        sigmaij = tf.sqrt(sigma[:, None] ** 2 + sigma[None, :] ** 2)
    if damping == 'fermi,dip':
        dipmat = damping_fermi(dists, S_vdw, a)[:, :, None, None] * T_bare(Rs)
    elif damping == 'fermi,dip,gg':
        dipmat = (1 - damping_fermi(dists, S_vdw, a)[:, :, None, None]) * T_erf_coulomb(
            Rs, sigmaij
        )
    else:
        raise ValueError(f'Unsupported damping: {damping}')
    n_atoms = tf.shape(coords)[0]
    return tf.reshape(tf.transpose(dipmat, (0, 2, 1, 3)), (3 * n_atoms, 3 * n_atoms))


def damping_fermi(R, S_vdw, d):
    return 1 / (1 + tf.exp(-d * (R / S_vdw - 1)))


def T_bare(R):
    R_2 = tf.reduce_sum(R**2, -1)
    # 1e10 rather than inf is necessary to avoid nan gradients
    R_5 = tf.sqrt(_set_diag(R_2, 1e10)) ** 5
    return (
        -3 * R[:, :, :, None] * R[:, :, None, :]
        + R_2[:, :, None, None] * np.eye(3)[None, None, :, :]
    ) / R_5[:, :, None, None]


def T_erf_coulomb(R, sigma):
    bare = T_bare(R)
    # 1e-10 rather than 0 is necessary to avoid nan gradients from sqrt(0)
    R_1 = tf.sqrt(_set_diag(tf.reduce_sum(R**2, -1), 1e-10))
    # 1e10 rather than inf is necessary to avoid nan gradients
    R_5 = _set_diag(R_1**5, 1e10)
    RR_R5 = R[:, :, :, None] * R[:, :, None, :] / R_5[:, :, None, None]
    zeta = R_1 / sigma
    theta = 2 * zeta / tf.sqrt(pi) * tf.exp(-(zeta**2))
    erf_theta = tf.erf(zeta) - theta
    return (
        erf_theta[:, :, None, None] * bare
        + (2 * (zeta**2) * theta)[:, :, None, None] * RR_R5
    )


def _set_diag(A, val):
    return tf.matrix_set_diag(A, tf.fill(tf.shape(A)[0:1], tf.cast(val, tf.float64)))


def _repeat(a, n):
    return tf.reshape(tf.tile(a[:, None], (1, n)), (-1,))



import re

import pkg_resources

from .pymbd import ang, from_volumes, mbd_energy, mbd_energy_species, screening

__all__ = ['mbd_energy', 'mbd_energy_species', 'screening', 'ang', 'from_volumes']
try:
    __version__ = pkg_resources.get_distribution('pymbd').version
    __version__ = re.split('[.-]', __version__, maxsplit=3)
    __version__ = (*map(int, __version__[:3]), *__version__[3:])
except pkg_resources.DistributionNotFound:
    __version__ = None
